@@ -22,9 +22,21 @@
 #ifndef _I2CBUSSES_H
 #define _I2CBUSSES_H
 
-void print_i2c_busses(int procfmt);
+struct i2c_adap {
+	int nr;
+	char *name;
+	const char *funcs;
+	const char *algo;
+};
 
+struct i2c_adap *gather_i2c_busses(void);
+void free_adapters(struct i2c_adap *adapters);
+
+int lookup_i2c_bus(const char *i2cbus_arg);
+int parse_i2c_address(const char *address_arg);
 int open_i2c_dev(const int i2cbus, char *filename, const int quiet);
 int set_slave_addr(int file, int address, int force);
 
+#define MISSING_FUNC_FMT	"Error: Adapter does not have %s capability\n"
+
 #endif
